import random
from datetime import datetime, timedelta
from typing import Any, Dict, List

import numpy as np
from pytz import timezone

from sktmls.models import MLSLightGBMModel, MLSModelError

TZ = timezone("Asia/Seoul")
PRODUCTS = [
    {"id": "NA00006535", "name": "T플랜 안심2.5G", "price": 43000, "seg": ["tplan", "young"]},
    {"id": "NA00006536", "name": "T플랜 안심4G", "price": 50000, "seg": ["tplan"]},
    {"id": "NA00006537", "name": "T플랜 에센스", "price": 69000, "seg": ["tplan"]},
    {"id": "NA00006538", "name": "T플랜 스페셜", "price": 79000, "seg": ["tplan", "young"]},
    {"id": "NA00006539", "name": "T플랜 맥스", "price": 100000, "seg": ["tplan", "young"]},
    {"id": "NA00006155", "name": "0플랜 스몰", "price": 33000, "seg": ["young"]},
    {"id": "NA00006156", "name": "0플랜 미디엄", "price": 50000, "seg": ["young"]},
    {"id": "NA00006157", "name": "0플랜 라지", "price": 69000, "seg": ["young"]},
    {"id": "NA00006794", "name": "T플랜 시니어 안심 2.8G", "price": 43000, "seg": ["senior"]},
    {"id": "NA00006795", "name": "T플랜 시니어 안심 4.5G", "price": 50000, "seg": ["senior"]},
    {"id": "NA00006796", "name": "T플랜 시니어 에센스", "price": 69000, "seg": ["senior"]},
    {"id": "NA00006797", "name": "T플랜 시니어 스페셜", "price": 79000, "seg": ["senior"]},
    {"id": "NA00006403", "name": "5GX 스탠다드", "price": 75000, "seg": ["5g"]},
    {"id": "NA00006404", "name": "5GX 프라임", "price": 89000, "seg": ["5g"]},
    {"id": "NA00006405", "name": "5GX 플래티넘", "price": 125000, "seg": ["5g"]},
]


class FeeNoEquipSingleEmbModel(MLSLightGBMModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 LightGBM 기반 클래스입니다.

    일반 요금제 추천 모델에 특화된 모델이며, 하나의 임베딩 피쳐를 사용합니다.
    """

    def __init__(
        self,
        model,
        model_name: str,
        model_version: str,
        model_features: List[str],
        default_feature_values: List[Any],
        default_context_value: str,
        emb_feature_indices: List[int] = None,
        conversion_formulas: Dict[str, Dict[str, Any]] = dict(),
        context_features: List[str] = [],
        context_values: List[Any] = [],
        product_context_indices: Dict[str, List[int]] = dict(),
    ):
        """
        ## Args

        - model: LightGBM으로 학습한 모델 객체
        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - model_features: (list(str)) 피쳐 리스트 (필수 피쳐: `age`, `bas_fee_amt`, `filter_five_g`)
        - default_feature_values: (list) 피쳐 기본값 리스트
        - default_context_value: (str) 기본 컨텍스트 값
        - emb_feature_indices: (optional) (list(int)) 임베딩 피쳐 중 선택할 인덱스 리스트 (기본값: None. 전체 임베딩 피쳐를 사용)
        - conversion_formulas: (optional) (dict(dict)) Feature conversion에 사용할 조건 dict (기본값: {})
        - context_features: (optional) (list(str)) 컨텍스트 생성을 위한 피쳐 리스트 (기본값: [])
        - context_values: (optional) (list) 컨텍스트 값 리스트 (기본값: [])
        - product_context_indices: (optional) (dict) 추천 상품 별 사용 컨텍스트 인덱스 맵 (기본값: {})

        ## Example

        ```python
        fee_no_equip_model_v20200904 = FeeNoEquipSingleEmbModel(
            model=lightgbm_model,
            model_name="fee_no_equip_model",
            model_version="v20200904",
            model_features=["feature1", "feature2", "feature3", "feature4", "emb_feature_list"],
            default_feature_values=[3.0, "Y", 10.0, "S", [0.1, 0.2, 0.3, 0.4]],
            default_context_value="context_default",
            emb_feature_indices=[0, 2],
            conversion_formulas={
                "feature4": {"map": {"S": 1, "D": 2}, "default": 0},
                "feature1": {"conditions": [[">", 0], ["<", 100]], "default": -1}
            },
            context_features=["filter_tfamilymoa", "context_data_overuse", "app_use_traffic_music_ratio_median_yn"],
            context_values=["tfamilymoa_value", "overuse_value", "music_ratio_value"],
            product_context_indices={
                "NA00006537": [0, 1],
                "NA00006539": [0, 1, 2],
                "NA00006817": [1]
            }
        )

        result = fee_no_equip_model_v20200904.predict([0.1, "N", 100, "S", [1, 2, 3, 4]])
        ```
        """
        assert len(model_features) == len(
            default_feature_values
        ), "`model_features`와 `default_feature_values`의 길이가 다릅니다."
        assert None not in default_feature_values, "`default_feature_values`에 None이 포함되어 있습니다."
        assert isinstance(default_feature_values[-1], list), "임베딩 피쳐는 리스트 형식이어야 합니다."
        assert default_feature_values[-1], "임베딩 피쳐의 기본값 리스트가 유효하지 않습니다."

        if emb_feature_indices is not None:
            assert isinstance(emb_feature_indices, list), "`emb_feature_indices`는 리스트 형식이어야 합니다."

        assert isinstance(conversion_formulas, dict), "`conversion_formulas`는 dict 형식이어야 합니다."
        for k, v in conversion_formulas.items():
            assert isinstance(k, str), "`conversion_formulas`의 키는 문자열이어야 합니다."
            assert isinstance(v, dict), "`conversion_formulas`의 값은 dict 형식이어야 합니다."
            assert "map" in v or "conditions" in v, "`conversion_formulas`의 값에 map 또는 conditions 키가 포함되어야 합니다."
            assert "default" in v, "`conversion_formulas`의 값에 default 키가 포함되어야 합니다."

        if context_features and context_values:
            assert len(context_features) == len(context_values), "`context_features`와 `context_values`의 길이는 같아야 합니다."
        else:
            assert (
                not context_features and not context_values
            ), "`context_features`와 `context_values`는 둘 다 전달되거나 둘 다 전달되지 않아야 합니다."

        assert isinstance(product_context_indices, dict), "`product_context_indices`는 dict 형식이어야 합니다."
        product_ids = [p["id"] for p in PRODUCTS]
        for k, v in product_context_indices.items():
            assert isinstance(k, str), "`product_context_indices`의 키는 문자열이어야 합니다."
            assert k in product_ids, "`product_context_indices`에 없는 상품 ID가 포함되어 있습니다."
            assert isinstance(v, list), "`product_context_indices`의 값은 리스트 형식이어야 합니다."
            assert len(v) <= len(context_features), "`product_context_indices`의 리스트가 `context_features`보다 깁니다."

        super().__init__(model, model_name, model_version, model_features + context_features)

        self.default_feature_values = default_feature_values
        self.default_context_value = default_context_value
        self.emb_feature_indices = (
            emb_feature_indices if emb_feature_indices is not None else list(range(len(default_feature_values[-1])))
        )
        self.conversion_formulas = conversion_formulas
        self.context_values = context_values
        self.product_context_indices = product_context_indices

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        # x 체크
        if len(self.features) != len(x):
            raise MLSModelError("FeeNoEquipSingleEmbModel: `x`의 길이가 `features`의 길이와 다릅니다.")

        model_x = x[: len(x) - len(self.context_values)]
        context_x = x[len(x) - len(self.context_values) :]

        if model_x[-1] and not isinstance(model_x[-1], list):
            raise MLSModelError("FeeNoEquipSingleEmbModel: model_x의 마지막 element가 list 타입이 아닙니다.")
        if model_x[-1] and (len(model_x[-1]) != len(self.default_feature_values[-1])):
            raise MLSModelError("FeeNoEquipSingleEmbModel: model_x와 `default_feature_values`의 임베딩 피쳐 길이가 다릅니다.")
        if model_x[-1] and (len(model_x[-1]) < len(self.emb_feature_indices)):
            raise MLSModelError("FeeNoEquipSingleEmbModel: `emb_feature_indices`의 길이가 model_x의 임베딩 피쳐 길이보다 큽니다.")

        # Feature 전처리
        non_emb_x = [f if f is not None else self.default_feature_values[i] for i, f in enumerate(model_x[:-1])]
        if model_x[-1]:
            emb_x = [
                model_x[-1][i] if model_x[-1][i] is not None else self.default_feature_values[-1][i]
                for i in self.emb_feature_indices
            ]
        else:
            emb_x = [self.default_feature_values[-1][i] for i in self.emb_feature_indices]

        try:
            preprocessed_x = self._convert(non_emb_x) + emb_x
        except Exception:
            raise MLSModelError("FeeNoEquipSingleEmbModel: Feature conversion에 실패했습니다.")

        # Realtime feature 생성
        now = datetime.now(TZ)
        preprocessed_x.append(now.weekday() + 1)
        preprocessed_x.append(now.day / (now.replace(day=1) - timedelta(days=1)).day)

        # 상품 선택을 위한 feature 추출
        try:
            age = preprocessed_x[self.features.index("age")]
            bas_fee_amt = preprocessed_x[self.features.index("bas_fee_amt")]
            filter_five_g = preprocessed_x[self.features.index("filter_five_g")]
        except Exception:
            raise MLSModelError("FeeNoEquipSingleEmbModel: Context feature가 존재하지 않습니다.")
        products = self._get_products(age, filter_five_g)
        if not products:
            return {"items": []}

        # Prediction
        y = self.models[0].predict(np.array([preprocessed_x], dtype=np.float32)).flatten().tolist()[0]
        if y <= 0:
            return {"items": []}

        # 상품 선택
        pred_bas_fee_amt = bas_fee_amt + y * 1000
        rec_prod = None
        min_diff = 999999
        for p in products:
            diff = abs(pred_bas_fee_amt - p["price"])
            if diff > min_diff:
                break
            min_diff = diff
            rec_prod = p
        if rec_prod["price"] <= bas_fee_amt:
            return {"items": []}

        # Context 빌드
        context_indices = self.product_context_indices.get(rec_prod["id"])
        if context_indices:
            contexts = [self.context_values[i] for i in context_indices if context_x[i] == "Y"]
            context = random.choice(contexts) if contexts else self.default_context_value
        else:
            context = self.default_context_value

        return {
            "items": [
                {
                    "id": rec_prod["id"],
                    "name": rec_prod["name"],
                    "type": "fee_no_equip",
                    "props": {"context_id": context},
                }
            ]
        }

    def _get_products(self, age, filter_five_g):
        if 25 <= age < 65 and filter_five_g == 0:
            seg = "tplan"
        elif 19 <= age < 25 and filter_five_g == 0:
            seg = "young"
        elif 65 <= age <= 80 and filter_five_g == 0:
            seg = "senior"
        elif 19 <= age <= 80 and filter_five_g == 1:
            seg = "5g"
        else:
            return None

        prods = list(filter(lambda p: seg in p["seg"], PRODUCTS))
        prods = sorted(prods, key=lambda p: p["price"])
        return prods

    def _convert(self, features):
        converted_features = []
        for i, f in enumerate(features):
            feature_name = self.features[i]
            formula = self.conversion_formulas.get(feature_name)
            if formula:
                if "map" in formula:
                    converted_features.append(formula["map"].get(f, formula["default"]))
                elif "conditions" in formula:
                    conditions = formula["conditions"]
                    conditions_checked = True
                    for condition in conditions:
                        if (
                            (condition[0] == "==" and f != condition[1])
                            or (condition[0] == "!=" and f == condition[1])
                            or (condition[0] == "<" and f >= condition[1])
                            or (condition[0] == "<=" and f > condition[1])
                            or (condition[0] == ">" and f <= condition[1])
                            or (condition[0] == ">=" and f < condition[1])
                        ):
                            conditions_checked = False
                            break
                    converted_features.append(f if conditions_checked else formula["default"])
            else:
                converted_features.append({"Y": 1, "N": 0}.get(f, f))
        return converted_features



import random
from datetime import datetime, timedelta
from functools import reduce
from typing import Any, Dict, List

import numpy as np
from pytz import timezone

from sktmls.models import MLSGenericModel, MLSModelError
from sktmls.utils import LogicProcessor

TZ = timezone("Asia/Seoul")

logic_processor = LogicProcessor()


class GenericContextModel(MLSGenericModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 모델 기반의 클래스입니다.

    y 결과에 따라 적절한 상품 정보와 context_id를 선택하여 제공해야 하는 경우 사용 가능합니다.
    """

    def __init__(
        self,
        model,
        model_name: str,
        model_version: str,
        model_features: List[str],
        default_model_feature_values: List[Any],
        default_context_value: str,
        products: Dict[str, Any],
        product_selection_logic: Dict[str, Any],
        conversion_formulas: Dict[str, Dict[str, Any]] = dict(),
        emb_features: List[str] = [],
        default_emb_feature_values: List[List[Any]] = [],
        emb_feature_indices: Dict[str, List[int]] = [],
        context_features: List[str] = [],
        context_values: List[Any] = [],
        realtime_features: List[str] = [],
        include_y: bool = False,
        y_key: str = "score",
    ):
        """
        ## Args

        - model: ML 라이브러리로 학습한 모델 객체
        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - model_features: (list(str)) 피쳐 리스트
        - default_model_feature_values: (list) 피쳐 기본값 리스트
        - default_context_value: (str) 기본 컨텍스트 값
        - products: (dict) 추천 상품 정보. 형식은 아래 Example 참조
        - product_selection_logic: (dict) 계산된 y로 추천 상품을 선택하는 로직. 형식은 아래 Example 참조
        - conversion_formulas: (optional) (dict(dict)) Feature conversion에 사용할 조건 dict. 형식은 아래 Example 참조 (기본값: {})
        - emb_features: (optional) (list(str)) 임베딩 피쳐 이름 리스트 (기본값: [])
        - default_emb_feature_values: (optional) (list(list)) 임베딩 피쳐들의 기본값 리스트 (리스트의 리스트 형식) (기본값: [])
        - emb_feature_indices: (optional) (list(list)) 각 임베딩 피쳐의 세부 사용 피쳐 인덱스 리스트 (기본값: [])
        - context_features: (optional) (list(str)) 컨텍스트 생성을 위한 피쳐 리스트 (기본값: [])
        - context_values: (optional) (list(str)) 컨텍스트 값 리스트 (기본값: [])
        - realtime_features: (optional) (list(str)) 실시간 피쳐 리스트 (포함 가능 값: `weekday`|`day_ratio`) (기본값: [])
        - include_y: (optional) (bool) props에 y값을 포함할 지 여부 (기본값: False)
        - y_key: (optional) (str) `include_y`가 True인 경우 props에서 y값을 쓰기 위해 사용할 키 (기본값: `score`)

        ## Example

        ```python
        model_features = ["feature1", "feature2", "feature3", "feature4", "feature5", "feature6"]
        emb_features = ["embedding_vector1", "embedding_vector2"]
        context_features = ["context_feature1", "context_feature2"]

        default_model_feature_values = ["D", 0.0, "N/A", 0.0, 0.0, "S"]
        default_emb_feature_values = [[1, 2, 3], [4, 5, 6]]
        emb_feature_indices = [[0, 1, 2], [0, 2]]

        # 상품 정보
        # - products: 추천 상품 정보. Key는 상품의 고유 id
        #    - name
        #    - type
        #    - context_features: 컨텍스트로 사용할 feature 리스트. None이거나 비어있으면 default_context_value가 반환됨
        #    - random_context: 유효한 컨텍스트를 랜덤하게 뽑아서 반환할 지 context_features의 순서대로 우선순위를 주어 반환할 지 여부 (기본값: False)
        products = {
            "PRODUCT001": {
                "name": "상품001",
                "type": "상품타입",
                "context_features": ["context_feature1", "context_feature2"],
                "random_context": False,
            },
            "PRODUCT002": {
                "name": "상품002",
                "type": "상품타입",
                "context_features": ["context_feature1"],
                "random_context": False,
            },
            "PRODUCT003": {
                "name": "상품002",
                "type": "상품타입",
                "context_features": ["context_feature1", "context_feature2"],
                "random_context": True,
            },
        }

        # 상품 선택 로직
        # - product_selection_logic: ML라이브러리 predict 함수 계산을 통해 얻어낸 y값을 바탕으로 실제 상품을 선택하는 로직
        # - 자세한 문법은 sktmls/utils/logic.py 참조
        product_selection_logic = {
            "if": [
                # 계산한 y가 0보다 크고
                # 25 <= feature2 < 65 이며
                # feature3 == "N" 인 경우
                # 아래 if_closest_upper 확인
                {
                    "and": [
                        {">": [{"var": "y"}, 0]},
                        {">=": [{"var": "feature2"}, 25]},
                        {"<": [{"var": "feature2"}, 65]},
                        {"==": [{"var": "feature3"}, "N"]},
                    ]
                },
                {
                    "if_closest_upper": [
                        # y * 1000 + feature4 값이 가장 가까운 것을 리턴
                        {"+": [{"*": [{"var": "y"}, 1000]}, {"var": "feature4"}]},
                        # 43000에 가장 가까운 경우
                        # "PRODUCT001"를 리턴
                        43000,
                        "PRODUCT001",
                        # 50000 가장 가까운 경우
                        # "PRODUCT002"를 리턴
                        50000,
                        "PRODUCT002",
                    ]
                },
                # 계산한 y가 0보다 크고
                # 19 <= feature2 < 25 이며
                # feature3 == "N" 인 경우
                # 아래 if_closest_upper 확인
                {
                    "and": [
                        {">": [{"var": "y"}, 0]},
                        {">=": [{"var": "feature2"}, 19]},
                        {"<": [{"var": "feature2"}, 25]},
                        {"==": [{"var": "feature3"}, "N"]},
                    ]
                },
                {
                    "if_closest_upper": [
                        # y * 1000 + feature4 값이 가장 가까운 것을 리턴
                        {"+": [{"*": [{"var": "y"}, 1000]}, {"var": "feature4"}]},
                        # 53000에 가장 가까운 경우
                        # "PRODUCT001"를 리턴
                        53000,
                        "PRODUCT001",
                        # 60000 가장 가까운 경우
                        # "PRODUCT003"를 리턴
                        60000,
                        "PRODUCT003",
                    ]
                },
                # 아무 조건에도 해당되지 않을 때 None
                None
            ]
        }

        conversion_formulas = {
            "feature1": {"map": {"D": 1, "V": 2}, "default": 0},
            "feature5": {"map": {"S": 1, "K": 2}, "default": 0},
        }

        realtime_features = ["weekday", "day_ratio"]

        my_model_v1 = GenericContextModel(
            model=model,
            model_name="my_model",
            model_version="v1",
            model_features=model_features,
            default_model_feature_values=default_model_feature_values,
            default_context_value="default_context_id",
            products=products,
            product_selection_logic=product_selection_logic,
            conversion_formulas=conversion_formulas,
            emb_features=emb_features,
            default_emb_feature_values=default_emb_feature_values,
            emb_feature_indices=emb_feature_indices,
            context_features=context_features,
            context_values=context_features,
            realtime_features=realtime_features
        )
        ```
        """
        assert len(model_features) == len(
            default_model_feature_values
        ), "`model_features`와 `default_model_feature_values`의 길이가 다릅니다."
        assert None not in default_model_feature_values, "`default_model_feature_values`에 None이 포함되어 있습니다."

        assert isinstance(emb_features, list), "`emb_features`는 리스트 형식이어야 합니다."
        assert isinstance(default_emb_feature_values, list), "`default_emb_feature_values`는 리스트 형식이어야 합니다."
        assert None not in default_emb_feature_values, "`default_emb_feature_values`에 None이 포함되어 있습니다."
        assert isinstance(emb_feature_indices, list), "`emb_feature_indices`는 리스트 형식이어야 합니다."
        assert None not in emb_feature_indices, "`emb_feature_indices`에 None이 포함되어 있습니다."
        assert (
            len(emb_features) == len(default_emb_feature_values) == len(emb_feature_indices)
        ), "`emb_features`, `default_emb_feature_values`, `emb_feature_indices`의 길이가 다릅니다."

        for i, indices in enumerate(emb_feature_indices):
            assert isinstance(indices, list), "`emb_feature_indices` 내 모든 element는 리스트 형식이어야 합니다."
            assert len(indices) <= len(
                default_emb_feature_values[i]
            ), "`emb_feature_indices` 내 element 길이가 해당 임베딩 피쳐 길이보다 깁니다."
            assert None not in default_emb_feature_values[i], "`default_emb_feature_values` 내 element에 None이 포함되어 있습니다."
            for idx in indices:
                assert isinstance(idx, int), "`emb_feature_indices` 내 모든 인덱스 값은 int 형식이어야 합니다."
                assert idx < len(default_emb_feature_values[i]), "`emb_feature_indices` 내 인덱스가 임베딩 피쳐 전체 길이보다 깁니다."

        assert isinstance(context_features, list), "`context_features`는 리스트 형식이어야 합니다."
        if context_features:
            assert context_values and isinstance(context_values, list), "`context_values`는 리스트 형식이어야 합니다."
            assert len(context_features) == len(context_values), "`context_features`와 `context_values`의 길이가 다릅니다."
        else:
            assert not context_values, "`context_features`와 `context_values`의 길이가 다릅니다."

        assert isinstance(realtime_features, list), "`realtime_features`는 리스트 형식이어야 합니다."
        assert set(realtime_features).issubset({"weekday", "day_ratio"}), "현재 지원하는 실시간 피쳐는 `weekday`와 `day_ratio` 입니다."

        assert isinstance(conversion_formulas, dict), "`conversion_formulas`는 dict 형식이어야 합니다."
        for k, v in conversion_formulas.items():
            assert isinstance(k, str), "`conversion_formulas`의 키는 문자열이어야 합니다."
            assert isinstance(v, dict), "`conversion_formulas`의 값은 dict 형식이어야 합니다."
            assert "map" in v or "conditions" in v, "`conversion_formulas`의 값에 map 또는 conditions 키가 포함되어야 합니다."
            assert "default" in v, "`conversion_formulas`의 값에 default 키가 포함되어야 합니다."

        assert isinstance(product_selection_logic, dict), "`product_selection_logic`는 dict 형식이어야 합니다."
        for key in product_selection_logic.keys():
            assert key in ["var", "missing", "missing_some"] or key in logic_processor.operations

        product_context_features = set()
        for p in products.values():
            product_context_features.update(p.get("context_features", []))
        assert product_context_features.issubset(
            context_features
        ), "`products`에 명세된 context_features에 `context_features` 파라미터에 없는 피쳐가 존재합니다."

        super().__init__([model], model_name, model_version, model_features + emb_features + context_features)

        self.default_model_feature_values = default_model_feature_values
        self.default_context_value = default_context_value
        self.products = products
        self.product_selection_logic = product_selection_logic
        self.conversion_formulas = conversion_formulas
        self.default_emb_feature_values = default_emb_feature_values
        self.emb_feature_indices = emb_feature_indices
        self.context_values = context_values
        self.realtime_features = realtime_features
        self.include_y = include_y
        self.y_key = y_key

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if len(self.features) != len(x):
            raise MLSModelError("GenericContextModel: `x`의 길이가 `features`의 길이와 다릅니다.")

        # 1. 피쳐 전처리

        # - 1.1. 피쳐 나누기
        non_emb_x = x[: len(x) - len(self.context_values) - len(self.default_emb_feature_values)]
        emb_x = x[
            len(x) - len(self.context_values) - len(self.default_emb_feature_values) : len(x) - len(self.context_values)
        ]
        context_x = x[len(x) - len(self.context_values) :]

        for i, emb_f in enumerate(emb_x):
            if not emb_f:
                emb_x[i] = self.default_emb_feature_values[i]
                continue
            if len(emb_f) != len(self.default_emb_feature_values[i]):
                raise MLSModelError("GenericContextModel: emb_x와 `default_emb_feature_values`의 임베딩 피쳐 길이가 다릅니다.")
            if len(emb_f) < len(self.emb_feature_indices[i]):
                raise MLSModelError("GenericContextModel: `emb_feature_indices`의 임베딩 피쳐 인덱스 길이가 emb_x 임베딩 피쳐 길이보다 큽니다.")

        # - 1.2. None 피쳐 기본값 할당 및 임베딩 피쳐 펼치기
        non_emb_x = [f if f is not None else self.default_model_feature_values[i] for i, f in enumerate(non_emb_x)]
        emb_x = [
            [
                emb_x[i][j] if emb_x[i][j] is not None else self.default_emb_feature_values[i][j]
                for j in self.emb_feature_indices[i]
            ]
            for i in range(len(emb_x))
        ]
        emb_x = reduce(lambda a, b: a + b, emb_x, [])

        # - 1.3. conversion_formulas를 이용한 피쳐 맵핑 작업
        try:
            preprocessed_x = self._convert(non_emb_x) + emb_x
        except Exception as e:
            raise MLSModelError(f"GenericContextModel: Feature conversion에 실패했습니다. {e}")

        # - 1.4. Realtime feature 생성
        now = datetime.now(TZ)
        for realtime_feature in self.realtime_features:
            if realtime_feature == "weekday":
                preprocessed_x.append(now.weekday() + 1)
            else:
                preprocessed_x.append(now.day / (now.replace(day=1) - timedelta(days=1)).day)

        # 2. Prediction
        try:
            y = self.models[0].predict(np.array([preprocessed_x], dtype=np.float32)).flatten().tolist()[0]
        except Exception as e:
            raise MLSModelError(f"GenericContextModel: Prediction에 실패했습니다. {e}")

        # 3. 컨텍스트 빌드

        # - 3.1. 추천 상품 선택
        try:
            non_emb_features = self.features[
                : len(self.features) - len(self.context_values) - len(self.default_emb_feature_values)
            ]
            context_features = self.features[len(self.features) - len(self.context_values) :]

            feature_map = {name: non_emb_x[i] for i, name in enumerate(non_emb_features)}
            feature_map.update({name: context_x[i] for i, name in enumerate(context_features)})
            feature_map["y"] = y

            product_id = logic_processor.apply(self.product_selection_logic, feature_map)
            if not product_id:
                return {"items": []}

            product = self.products[product_id]
        except Exception as e:
            raise MLSModelError(f"GenericContextModel: 추천 상품 선택 로직 프로세스에 실패했습니다. {e}")

        # - 3.2. context_id 선택
        try:
            product_context_features = [f for f in product.get("context_features", []) if feature_map[f] == "Y"]
            if product_context_features:
                context_map = {name: self.context_values[i] for i, name in enumerate(context_features)}
                context_ids = [context_map[f] for f in product_context_features]
                is_random = product.get("random_context", False)
                context_id = random.choice(context_ids) if is_random else context_ids[0]
            else:
                context_id = self.default_context_value
        except Exception as e:
            raise MLSModelError(f"GenericContextModel: context_id 선택에 실패했습니다. {e}")

        # 4. return
        item = {
            "id": product_id,
            "name": product["name"],
            "type": product["type"],
            "props": {"context_id": context_id},
        }
        if self.include_y:
            item["props"][self.y_key] = round(y, 4)
        return {"items": [item]}

    def _convert(self, features):
        converted_features = []
        for i, f in enumerate(features):
            feature_name = self.features[i]
            formula = self.conversion_formulas.get(feature_name)
            if formula:
                if "map" in formula:
                    converted_features.append(formula["map"].get(f, formula["default"]))
                elif "conditions" in formula:
                    conditions = formula["conditions"]
                    conditions_checked = True
                    for condition in conditions:
                        if (
                            (condition[0] == "==" and f != condition[1])
                            or (condition[0] == "!=" and f == condition[1])
                            or (condition[0] == "<" and f >= condition[1])
                            or (condition[0] == "<=" and f > condition[1])
                            or (condition[0] == ">" and f <= condition[1])
                            or (condition[0] == ">=" and f < condition[1])
                        ):
                            conditions_checked = False
                            break
                    converted_features.append(f if conditions_checked else formula["default"])
            else:
                converted_features.append({"Y": 1, "N": 0}.get(f, f))
        return converted_features



from dateutil import parser
from enum import Enum
from string import Template
from typing import Any, List

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSResponse

SPARK_QUERY_FOR_LABLE_DATA = Template(
    """
SELECT sha2(source.svc_mgmt_num, 256) AS svc_mgmt_num, source.period AS period
FROM (
    ${source_query}
) AS source
"""
)


class ProblemType(Enum):
    """
    AutoML 데이터셋의 문제 타입입니다.

    - SCORE: 스코어 기반 고객 타겟팅
    - CLASSIFICATION: 분류
    - REGRESSION: 회귀
    """

    SCORE = "score"
    CLASSIFICATION = "classification"
    REGRESSION = "regression"


class DatasetError(Exception):
    def __init__(self, *args, **kwargs):
        """
        AutoML 데이터셋 에러.
        """
        super().__init__(*args, **kwargs)


class FeatureStoreConf:
    def __init__(self, enabled: bool, feature_group_id_list: List[int] = None, n_label_ratio: float = 1.0):
        """
        AutoML 데이터셋의 피쳐 스토어 설정입니다.

        ## Args

        - enabled: (bool) 피쳐 스토어 사용 유무
        - feature_group_id_list: (optional) (list(int)) 사용하고자 하는 피쳐 그룹의 ID 리스트
        - n_label_ratio: (optional) (float) N 레이블 샘플링 비율 (스코어 기반 고객 타겟팅 문제만 지원, 기본값: 1.0)

        ## Example

        ```python
        config = FeatureStoreConf(enabled=True, feature_group_id_list=[1, 2, 3], n_label_ratio=1.5)
        ```
        """
        self.enabled = enabled
        self.feature_group_id_list = feature_group_id_list
        self.n_label_ratio = n_label_ratio

    def __str__(self) -> str:
        return str(self.__dict__)


class LabelDataConf:
    def __init__(self, source_type: str, source_path: str):
        """
        AutoML 데이터셋의 레이블 데이터 설정입니다.

        ## Args

        - source_type: (str) 레이블 데이터 타입 (`csv` | `edd2` | `ye`).
        - source_path: (str) 레이블 데이터의 이름 또는 테이블 경로.
          - csv: 파일 이름 (ex. test.csv)
          - edd2: EDD -> MLS 로 전달된 BigQuery 테이블의 Full 경로 (ex. sktaic-datahub.mls_edd2_sync.test)
          - ye: YE -> MLS 로 전달된 데이터의 이름 (ex. test)

        ## Example

        ```python
        config = LabelDataConf(source_type="ye", source_path="my_test_table")
        ```
        """
        self.source_type = source_type
        self.source_path = source_path

    def __str__(self) -> str:
        return str(self.__dict__)


class Dataset:
    def __init__(self, **kwargs):
        """
        AutoML 학습 데이터셋 클래스.

        ## Attributes

        - id: (int) 데이터셋 ID
        - name: (str) 데이터셋 이름
        - status: (str) 데이터셋 상태
        - problem_type: (`sktmls.datasets.ProblemType`) 문제 타입
        - feature_store_conf: (`sktmls.datasets.FeatureStoreConf`) 피쳐 스토어 설정값
        - label_data_conf: (`sktmls.datasets.LabelDataConf`) 레이블 데이터 설정값
        - created_at: (datetime.datetime) 데이터셋 생성 시점
        - updated_at: (datetime.datetime) 데이터셋 생성 완료 또는 갱신 시점
        """
        self.id = kwargs.get("id")
        self.name = kwargs.get("name")
        self.status = kwargs.get("status")
        self.problem_type = ProblemType(kwargs.get("data_type"))

        base_feature_conf = kwargs.get("setup_conf", {}).get("base_feature_conf", {})
        self.feature_store_conf = FeatureStoreConf(
            enabled=base_feature_conf.get("enabled"),
            feature_group_id_list=base_feature_conf.get("feature_group_id_list"),
            n_label_ratio=base_feature_conf.get("n_label_ratio"),
        )

        label_data_conf = kwargs.get("setup_conf", {}).get("label_data_conf", {})
        self.label_data_conf = LabelDataConf(
            source_type=label_data_conf.get("source_type"),
            source_path=label_data_conf.get("source_path"),
        )

        try:
            self.created_at = parser.parse(kwargs.get("created_at"))
        except TypeError:
            self.created_at = None

        try:
            self.updated_at = parser.parse(kwargs.get("updated_at"))
        except TypeError:
            self.updated_at = None

    def __str__(self) -> str:
        return self.name


class DatasetClient(MLSClient):
    def __init__(
        self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, username: str = None, password: str = None
    ):
        """
        AutoML 데이터셋 관련 기능을 제공하는 클라이언트 클래스입니다.

        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        ## Returns
        `sktmls.datasets.DatasetClient`

        ## Example

        ```python
        client = DatasetClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def create_dataset(
        self, name: str, problem_type: ProblemType, feature_store_conf: FeatureStoreConf, label_data_conf: LabelDataConf
    ) -> MLSResponse:
        """
        새 AutoML 데이터셋을 생성합니다.

        ## Args

        - name: (str) 데이터셋 이름
        - problem_type: (`sktmls.datasets.ProblemType`) 문제 타입
        - feature_store_conf: (`sktmls.datasets.FeatureStoreConf`) 피쳐 스토어 설정
        - label_data_conf: (`sktmls.datasets.LabelDataConf`) 레이블 데이터 설정

        ## Returns
        `sktmls.datasets.DatasetClient`

        ## Example

        ```python
        client = DatasetClient()
        client.create_dataset(
            name="my_dataset",
            problem_type=ProblemType.SCORE,
            feature_store_conf=FeatureStoreConf(enabled=True, feature_group_id_list=[1, 2, 3], n_label_ratio=1.5),
            label_data_conf=LabelDataConf(source_type="ye", source_path="my_test_table"),
        )
        ```
        """
        data = {
            "name": name,
            "type": problem_type.value,
            "base_feature_conf": feature_store_conf.__dict__,
            "label_data_conf": label_data_conf.__dict__,
        }
        return self._request(method="POST", url="api/v1/datasets", data=data)

    def list_datasets(self, **kwargs) -> List[Dataset]:
        """
        AutoML 데이터셋 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 필터 조건
            - id: (int) 데이터셋 ID
            - name: (str) 데이터셋 이름
            - query: (str) 검색 문자
            - page: (int) 페이지 번호

        ## Returns
        list(`sktmls.datasets.Dataset`)
        """
        response = self._request(method="GET", url="api/v1/datasets", params=kwargs)

        if response.code != 0:
            raise DatasetError(f"Fail to search datasets: {response.error}")

        return [Dataset(**dataset) for dataset in response.results]

    def get_dataset(self, name: str = None, id: int = None) -> Dataset:
        """
        해당하는 AutoML 데이텨셋을 가져옵니다.

        ## Args: `id` 또는 `name` 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) 데이터셋 ID
        - name: (str) 데이터셋 이름

        ## Returns
        `sktmls.datasets.Dataset`
        """
        assert id is not None or name, "One of id or name must be provided"

        datasets = self.list_datasets(name=name, id=id)

        if len(datasets) == 0:
            raise DatasetError("Dataset does not exists")

        return datasets[0]

    def delete_dataset(self, name: str = None, id: int = None) -> None:
        """
        해당하는 AutoML 데이터셋을 삭제합니다

        ## Args: `id` 또는 `name` 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) 데이터셋 ID
        - name: (str) 데이터셋 이름

        """
        dataset = self.get_dataset(name=name, id=id)
        response = self._request(method="DELETE", url=f"api/v1/datasets/{dataset.id}")

        if response.code != 0:
            raise DatasetError(f"Fail to delete a dataset: {response.error}")

    def upload_label_data_from_ye(self, spark_session: Any, output_name: str, source_query: str) -> None:
        """
        YE 환경에서 레이블 데이터를 업로드합니다. 전달하려는 데이터의 원천 데이터는 Spark Query로 조회 가능해야 합니다.

        ## Args

        - spark_session: (SparkSession) `skt` 패키지의 skt.ye.get_spark() 함수를 통해 얻어지는 `SparkSession` 객체.
        - output_name: (str) 출력하고자 하는 데이터 이름.
        - source_query: (str) 업로드하고자 하는 데이터를 조회하기 위한 Spark Query(`svc_mgmt_num`, `period` 컬럼이 필수로 포함되어야 함).

        ## Example

        ```python
        source_query = '''
        SELECT svc_mgmt_num, '202001' AS period
        FROM test_db.test_table
        LIMIT 10000
        '''

        client = DatasetClient()
        client.upload_label_data_from_ye(spark_session=skt.ye.get_spark(), output_name="ye_test.csv", source_query=source_query)
        ```
        """
        assert type(spark_session).__name__ == "SparkSession", "Invalid type of spark_session"
        assert type(output_name) == str, "Invalid type of output_name"
        assert type(source_query) == str, "Invalid type of source_query"

        if not self.config.MLS_RUNTIME_ENV == "YE":
            raise DatasetError("upload_label_data_from_ye() is only supported over the YE environment.")

        query = SPARK_QUERY_FOR_LABLE_DATA.substitute(source_query=source_query)
        target_path = f"gs://sktmls-automl-{self.get_env().value}/{self.get_username()}/label_data_ye/{output_name}"
        spark_session.sql(query).write.mode("overwrite").csv(target_path, header=True)
        spark_session.stop()

    def upload_label_data_from_ye_bq(
        self,
        bigquery_client: Any,
        source_dataset: str,
        source_table: str,
        output_name: str = None,
    ) -> None:
        """
        YE 환경에서 BigQuery에 존재하는 테이블을 본인 MLS계정에 레이블 데이터로 업로드합니다.

        ## Args

        - bigquery_client: (google.cloud.bigquery.client.Client) `skt` 패키지의 skt.gcp.get_bigquery_client() 함수를 통해 얻어지는 `google.cloud.bigquery.client.Client` 객체.
        - source_dataset: (str) 업로드하고자 하는 BigQuery 데이터셋 이름.
        - source_table: (str) 업로드하고자 하는 BigQuery 테이블 이름 (`svc_mgmt_num`, `period` 컬럼이 필수로 포함되어야 함).
        - output_name: (optional) (str) 출력하고자 하는 데이터 이름. (기본값 : `source_table`)

        ## Example

        ```python
        client = DatasetClient()
        client.upload_label_data_from_ye_bq(bigquery_client=skt.gcp.get_bigquery_client(), source_dataset="my_dataset", source_table="my_table")
        ```
        """
        assert (
            str(type(bigquery_client)) == "<class 'google.cloud.bigquery.client.Client'>"
        ), "Invalid type of bigquery_client"
        assert type(source_dataset) == str, "Invalid type of source_dataset"
        assert type(source_table) == str, "Invalid type of source_table"
        if output_name is not None:
            type(output_name) == str, "Invalid type of otuput_name"

        if not output_name:
            output_name = source_table

        if not self.config.MLS_RUNTIME_ENV == "YE":
            raise DatasetError("upload_label_data_from_ye() is only supported over the YE environment.")

        try:
            bigquery_table = bigquery_client.get_table(f"sktaic-datahub.{source_dataset}.{source_table}")
        except Exception as e:
            raise DatasetError(str(e))

        try:
            extract_job = bigquery_client.extract_table(
                source=bigquery_table,
                destination_uris=f"gs://sktmls-automl-{self.get_env().value}/{self.get_username()}/label_data_ye/{output_name}/part-{output_name}",
            )
            extract_job.result()
        except Exception as e:
            raise DatasetError(str(e))



import pytest

import numpy as np

import lightgbm

from sktmls.models.contrib import GenericContextModel


class TestLightGBMRandomContextModels:
    @pytest.fixture(scope="class")
    def param(self):
        return {
            "model": GenericContextModel(
                model=lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
                model_name="test_model",
                model_version="test_version",
                model_features=["feature1", "feature2", "feature3", "feature4"],
                default_model_feature_values=[0.0, "N", 10.0, "S"],
                default_context_value="default_context",
                products={
                    "PROD001": {
                        "name": "제품1",
                        "type": "my_type",
                        "context_features": ["cx_feat1", "cx_feat2"],
                        "random_context": False,
                    },
                    "PROD002": {
                        "name": "제품2",
                        "type": "my_type",
                        "context_features": ["cx_feat1"],
                        "random_context": False,
                    },
                },
                product_selection_logic={
                    "if": [
                        {"and": [{">": [{"var": "y"}, 0.5]}, {"<=": [{"var": "feature3"}, 5]}]},
                        "PROD001",
                        {"<=": [{"var": "y"}, 0.5]},
                        "PROD002",
                        None,
                    ]
                },
                conversion_formulas={"feature4": {"map": {"S": 0, "L": 1}, "default": -1}},
                emb_features=["emb1", "emb2"],
                default_emb_feature_values=[[0.0, 0.0, 0.0, 0.0], [1.0, 1.0, 1.0, 1.0]],
                emb_feature_indices=[[0, 1], [0, 2, 3]],
                context_features=["cx_feat1", "cx_feat2", "cx_feat3"],
                context_values=["cx_value1", "cx_value2", "cx_value3"],
                realtime_features=[],
                include_y=True,
                y_key="test_score",
            )
        }

    def test_000_generic_context_success(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.6]))
        assert param.get("model").predict([2.0, "Y", 4.0, "S", [1, 2, 3, 4], [10, 9, 8, 7], "Y", "Y", "Y"]) == {
            "items": [
                {
                    "id": "PROD001",
                    "name": "제품1",
                    "type": "my_type",
                    "props": {"context_id": "cx_value1", "test_score": 0.6},
                }
            ]
        }

    def test_001_generic_context_success(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.4]))
        assert param.get("model").predict([2.0, "Y", 4.0, "S", [1, 2, 3, 4], [10, 9, 8, 7], "N", "Y", "Y"]) == {
            "items": [
                {
                    "id": "PROD002",
                    "name": "제품2",
                    "type": "my_type",
                    "props": {"context_id": "default_context", "test_score": 0.4},
                }
            ]
        }

    def test_002_generic_context_selection_none(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.7]))
        assert param.get("model").predict([2.0, "Y", 10.0, "S", [1, 2, 3, 4], [10, 9, 8, 7], "Y", "Y", "Y"]) == {
            "items": []
        }



from .mls_model import (
    MLSModel,
    MLSLightGBMModel,
    MLSXGBoostModel,
    MLSCatBoostModel,
    MLSRuleModel,
    MLSGenericModel,
    MLSModelError,
)
from .ml_model import MLModelClient, MLModel, AutoMLModel, ManualModel, MLModelStatus

__all__ = [
    "MLSModel",
    "MLSLightGBMModel",
    "MLSXGBoostModel",
    "MLSCatBoostModel",
    "MLSRuleModel",
    "MLSGenericModel",
    "MLSModelError",
    "MLModelClient",
    "MLModel",
    "AutoMLModel",
    "ManualModel",
    "MLModelStatus",
]



import pytest

from sktmls.models import MLSModelError
from sktmls.models.contrib import MbrCategoryBaselineModel


class TestMbrCategoryBaselineModel:
    def test_000_mbr_baseline_1(self):
        mbm = MbrCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_benefit_pizza", "used_datediff_pizza"],
            item_type="test_type",
            num_pick=1,
        )

        results = mbm.predict(["1234567890", ["2234", "2308", "2306", "2286"], 37])
        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] in ["2234", "2308", "2306", "2286"]
        assert item["name"] == "#"
        assert item["type"] == "test_type"
        assert item["props"]["category_use_yn"] == "Y"

    def test_001_mbr_baseline_2(self):
        mbm = MbrCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_benefit_pizza", "used_datediff_pizza"],
            item_type="test_type",
            num_pick=2,
        )

        results = mbm.predict(["1234567890", ["2234", "2308", "2306", "2286"], 37])
        assert len(results["items"]) == 2

        item = results["items"][0]
        assert item["id"] in ["2234", "2308", "2306", "2286"]
        assert item["name"] == "#"
        assert item["type"] == "test_type"
        assert item["props"]["category_use_yn"] == "Y"

        item = results["items"][1]
        assert item["id"] in ["2234", "2308", "2306", "2286"]
        assert item["name"] == "#"
        assert item["type"] == "test_type"
        assert item["props"]["category_use_yn"] == "Y"

    def test_002_mbr_baseline_false_use_yn(self):
        mbm = MbrCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_benefit_pizza", "used_datediff_pizza"],
            item_type="test_type",
            num_pick=1,
        )

        results = mbm.predict(["1234567890", ["2234", "2308", "2306", "2286"], 1000])
        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] in ["2234", "2308", "2306", "2286"]
        assert item["name"] == "#"
        assert item["type"] == "test_type"
        assert item["props"]["category_use_yn"] == "N"

    def test_003_mbr_baseline_none_benefit(self):
        mbm = MbrCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_benefit_pizza", "used_datediff_pizza"],
            item_type="test_type",
            num_pick=1,
        )

        results = mbm.predict(["1234567890", None, 37])
        assert len(results["items"]) == 0

    def test_004_mbr_baseline_invalid_benefit(self):
        mbm = MbrCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_benefit_pizza", "used_datediff_pizza"],
            item_type="test_type",
            num_pick=1,
        )

        with pytest.raises(MLSModelError):
            mbm.predict(["1234567890", 100, 37])



import re

from dateutil import parser
from typing import Union

from sktmls import MLSClient, MLSRuntimeENV
from sktmls.datasets import DatasetClient, ProblemType
from sktmls.models import AutoMLModel


class AutoMLPredictionError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ScoreExportOption:
    def __init__(self, sort: bool, target_limit: int = None):
        """
        스코어 출력 옵션 값입니다(스코어 기반 고객 타겟팅 문제만 지원).

        ## Args

        - sort: (bool) 결과 정렬 여부
        - target_limit: (int) 총 출력 개수 (sort = True 인 경우 필수)

        ## Returns
        `sktmls.models.automl.ScoreExportOption`

        ## Example

        ```python
        score_export_option = ScoreExportOption(sort=True, target_limit=1000000)
        ```
        """
        assert type(sort) == bool, "sort 은(는) bool 타입이어야 합니다."
        assert type(target_limit) == int if target_limit is not None else True, "target_limit 은(는) int 타입이어야 합니다."
        assert target_limit is not None if sort else True, "sort = True 인 경우 target_limit 은 반드시 전달되어야 합니다."
        assert target_limit <= 10000000 if sort else True, "target_limit 값은 10,000,000 이하여야 합니다."
        self.sort = sort
        self.target_limit = target_limit


class CustomFeatureSourceConfig:
    def __init__(self, source_type: str, source_path: str):
        """
        예측 시 추가로 사용할 사용자 피쳐 데이터 설정입니다(사용자 피쳐를 추가로 적용한 모델의 경우 필수).

        ## Args

        - source_type: (str) 데이터 종류 (`csv` | `edd2` | `ye`)
        - source_path: (str) 데이터의 이름(또는 경로)
        """
        assert source_type in ["csv", "edd2", "ye"], "source_type 은(는) ['csv', 'edd2', 'ye'] 중 하나여야 합니다."
        assert type(source_path) == str, "source_path 은(는) str 타입이어야 합니다."
        self.source_type = source_type
        self.source_path = source_path


class ExportChannelConfig:
    def __init__(self, channel: str):
        self.channel = channel


class EmailChannelConfig(ExportChannelConfig):
    param_name = "email_channel_config"

    def __init__(self, email: str):
        """
        이메일로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - email: (str) 전송받을 이메일 경로 (@sktelecom.com 또는 @sk.com 도메인만 지원)

        ## Returns
        `sktmls.models.automl.EmailChannelConfig`

        ## Example

        ```python
        config = EmailChannelConfig("test@sktelecom.com")
        ```
        """
        super().__init__("EMAIL")
        assert type(email) == str, "email 은(는) str 타입이어야 합니다."
        self.email = email


class TOSChannelConfig(ExportChannelConfig):
    param_name = "tos_channel_config"

    def __init__(self, employee_num: str, target_group_name: str):
        """
        TOS로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - employee_num: (str) 사번 (TOS 계정)
        - target_group_name: (str) 원하는 대상군 명

        ## Returns
        `sktmls.models.automl.TOSChannelConfig`

        ## Example

        ```python
        config = TOSChannelConfig("test@sktelecom.com")
        ```
        """
        super().__init__("TOS")
        assert type(employee_num) == str, "employee_num 은(는) str 타입이어야 합니다."
        assert type(target_group_name) == str, "target_group_name 은(는) str 타입이어야 합니다."
        self.employee_num = employee_num
        self.target_group_name = target_group_name


class EDDChannelConfig(ExportChannelConfig):
    param_name = "edd_channel_config"

    def __init__(self, username: str, database: str, table: str):
        """
        EDD로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - username: (str) EDD 계정 (ex. g1234567)
        - database: (str) 데이터베이스 이름
        - table: (str) 테이블 이름

        ## Returns
        `sktmls.models.automl.EDDChannelConfig`

        ## Example

        ```python
        config = EDDChannelConfig("g1234567", "test_db", "test_table")
        ```
        """
        super().__init__("EDD")
        assert type(username) == str, "username 은(는) str 타입이어야 합니다."
        assert type(database) == str, "database 은(는) str 타입이어야 합니다."
        assert type(table) == str, "table 은(는) str 타입이어야 합니다."
        assert re.match(r"^[a-z][0-9]{7}", username), "잘못된 형식의 EDD 계정입니다."
        assert re.match(r"^[A-Za-z0-9_]+$", database), "잘못된 형식의 데이터베이스 이름입니다."
        assert re.match(r"^[A-Za-z0-9_]+$", table), "잘못된 형식의 테이블 이름입니다."
        self.username = username
        self.database = database
        self.table = table


class YEChannelConfig(ExportChannelConfig):
    param_name = "ye_channel_config"

    def __init__(self, username: str, table: str):
        """
        YE로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - username: (str) YE 계정 (ex. x1234567)
        - table: (str) 테이블 이름

        ## Returns
        `sktmls.models.automl.YEChannelConfig`

        ## Example

        ```python
        config = YEChannelConfig("x1234567", "test_table")
        ```
        """
        super().__init__("YE")
        assert type(username) == str, "username 은(는) str 타입이어야 합니다."
        assert type(table) == str, "table 은(는) str 타입이어야 합니다."
        assert re.match(r"^x[0-9]{7}", username), "잘못된 형식의 YE 계정입니다."
        assert re.match(r"^[A-Za-z0-9_]+$", table), "잘못된 형식의 테이블 이름입니다."
        self.username = username
        self.table = table


class AutoMLBatchPrediction:
    def __init__(self, **kwargs):
        """
        AutoML 배치 예측 작업 클래스입니다.

        ## Args

        - id: (int) 예측 작업 ID.
        - channel: (string) 전송 채널 (`EMAIL` | `TOS` | `EDD` | `YE`)
        - status: (string) 예측 진행 상태 (`RUNNING` | `DONE` | `ERROR_BATCH_PREDICT`)
        - created_at: (datetime) 예측 시작 시간
        - updated_at: (datetime) 예측 완료 시간

        ## Returns
        `sktmls.models.automl.AutoMLBatchPrediction`
        """
        self.id = kwargs.get("id")
        self.channel = kwargs.get("name")
        self.status = kwargs.get("status")

        try:
            self.created_at = parser.parse(kwargs.get("created_at"))
        except TypeError:
            self.created_at = None

        try:
            self.updated_at = parser.parse(kwargs.get("updated_at"))
        except TypeError:
            self.updated_at = None

    def __str__(self) -> str:
        return str(self.id)


class AutoMLBatchPredictionClient(MLSClient):
    def __init__(self, env=None, runtime_env: MLSRuntimeENV = None, username=None, password=None):
        """
        AutoML 배치 예측 관련 기능을 제공하는 클라이언트입니다.

        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.models.automl.AutoMLBatchPredictionClient`

        ## Example

        ```python
        client = AutoMLBatchPredictionClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def predict(
        self,
        model: AutoMLModel,
        export_channel_config: Union[EmailChannelConfig, TOSChannelConfig, EDDChannelConfig, YEChannelConfig],
        score_export_option: ScoreExportOption = None,
        period: str = None,
        custom_feature_source_config: CustomFeatureSourceConfig = None,
    ) -> AutoMLBatchPrediction:
        """
        배치 예측을 수행합니다.

        ## Args

        - model: (`sktmls.models.AutoMLModel`) 학습 모델
        - export_channel_config: (`sktmls.models.automl.ExportChannelConfig`) 예측 결과 전송 채널 설정
        - score_export_option: (`sktmls.models.automl.ScoreExportOption`) 스코어 전송 옵션 (스코어 기반 고객 타겟팅 문제의 경우 필수)
        - period: (str) 예측에 적용할 피쳐 스토어 기준연월 (형식: YYYYMM, 만약 입력하지 않을 경우 가장 최신의 기준연월을 적용)
        - custom_feature_source_config: (`sktmls.models.automl.CustomFeatureSourceConfig`) 예측에 적용할 커스텀 피쳐 데이터 파일 설정 (커스텀 피쳐를 사용한 모델의 경우 필수)

        ## Returns
        `sktmls.models.automl.AutoMLBatchPrediction`

        ## Example

        ```python
        client = AutoMLBatchPredictionClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")

        model = model_client.get_automl_model(id=72)
        export_channel_config = YEChannelConfig(username="x1234567", table="triple_a_test")
        score_export_option = ScoreExportOption(sort=True, target_limit=100)
        custom_feature_config = CustomFeatureSourceConfig("ye", "my_custom_feature_table")

        prediction = client.predict(
            model=model,
            export_channel_config=export_channel_config,
            score_export_option=score_export_option,
            custom_feature_source_config=custom_feature_config,
        )
        ```
        """
        assert type(model) == AutoMLModel, "model 은(는) AutoMLModel 타입이어야 합니다."
        assert type(export_channel_config) in [
            EmailChannelConfig,
            TOSChannelConfig,
            EDDChannelConfig,
            YEChannelConfig,
        ], "export_channel_config 은(는) [EmailChannelConfig, TOSChannelConfig, EDDChannelConfig, YEChannelConfig] 중 하나여야 합니다."
        assert (
            score_export_option is None or type(score_export_option) == ScoreExportOption
        ), "score_export_option 은(는) ScoreExportOption 타입이어야 합니다."
        assert period is None or type(period) == str, "period 은(는) str 타입이어야 합니다."

        dataset_client = DatasetClient(
            self._MLSClient__env, self._MLSClient__runtime_env, self._MLSClient__username, self._MLSClient__password
        )
        dataset = dataset_client.get_dataset(id=model.dataset_id)

        if dataset.problem_type == ProblemType.SCORE and not score_export_option:
            raise AutoMLPredictionError("스코어 기반 고객 타겟팅 문제의 경우 score_export_option 파라미터가 필수입니다.")

        export_config = export_channel_config.__dict__
        channel = export_channel_config.channel
        del export_config["channel"]

        data = {"channel": channel}
        data[export_channel_config.param_name] = export_config

        if score_export_option:
            data["score_export_option"] = score_export_option.__dict__

        if period:
            data["period"] = period

        if custom_feature_source_config:
            data["custom_feature_source_config"] = custom_feature_source_config.__dict__

        return AutoMLBatchPrediction(
            **self._request(method="POST", url=f"/api/v1/models/{model.id}/predictions", data=data).results
        )



from typing import List

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSResponse, MLSClientError

MLS_PROFILES_API_URL = "/api/v1/profiles"


class Profile:
    """
    MLS 프로파일 클래스입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) 프로파일 고유 ID
            - api_type: (str) 프로파일 타입 (`user`|`item`)
            - profile_id: (str) 프로파일 이름
            - tables: (list) 프로파일에 연결할 DynamoDB 테이블 이름 리스트
            - user: (str) 프로파일 소유 계정명

        ## Returns
        `sktmls.profiles.Profile`
        """
        self.id = kwargs.get("id")
        self.api_type = kwargs.get("api_type")
        self.profile_id = kwargs.get("profile_id")
        self.tables = kwargs.get("tables")
        self.user = kwargs.get("user")

    def get(self):
        return self.__dict__

    def reset(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class ProfileClient(MLSClient):
    """
    MLS 프로파일 관련 기능들을 제공하는 클라이언트입니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
        username: str = None,
        password: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        ## Returns
        `sktmls.profiles.ProfileClient`

        ## Example

        ```python
        profile_client = ProfileClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def create_profile(self, api_type: str, profile_id: str, tables: List[str], main_table: str = None) -> Profile:
        """
        프로파일을 생성합니다.

        ## Args

        - api_type: (str) 프로파일 타입 (`user`|`item`)
        - profile_id: (str) 프로파일 이름
        - tables: (list) 프로파일에 연결할 DynamoDB 테이블 이름 리스트
            - api_type이 `user`인 경우 모든 테이블 이름은 `user-profile`로 시작해야 합니다.
            - api_type이 `item`인 경우 모든 테이블 이름은 `item-profile`로 시작해야 합니다.
            - 테이블 간 중복 키에 대한 값 우선 순위는 왼쪽 테이블이 오른쪽 테이블보다 높습니다.
            - 연결된 모든 테이블 내에 없는 키에 대한 값은 null로 반환됩니다.
        - main_table: (optional) (str) 서비스 대상 여부를 결정하는 테이블
            - 기준 테이블에 user_id/item_id가 존재하지 않는 경우 다른 테이블에 데이터가 있다 하더라도 서비스되지 않습니다.
            - 기준 테이블에 user_id/item_id가 존재하는 경우 다른 테이블의 데이터 여부와 관계 없이 반드시 서비스됩니다.
            - 미선택 시 `tables` 파라미터에 명시된 테이블 중 첫 번째 테이블로 설정됩니다.

        ## Returns
        `sktmls.profiles.Profile`

        - id: (int) 프로파일 고유 ID
        - api_type: (str) 프로파일 타입 (`user`|`item`)
        - profile_id: (str) 프로파일 이름
        - tables: (list) 프로파일에 연결할 DynamoDB 테이블 이름 리스트
        - main_table: (str) 연결 테이블 중 기준이 되는 테이블 (추후 로직 추가 예정)
        - user: (str) 프로파일 소유 계정명

        ## Example

        ```python
        profile = profile_client.create_profile(
            api_type="user",
            profile_id="my_profile",
            tables=["user-profile", "user-profile-daily"]
        )
        ```
        """
        assert api_type in ["user", "item"], "api_type은 `user` 또는 `item`이어야 합니다."
        assert type(tables) == list
        for table_name in tables:
            assert type(table_name) == str, "`tables`는 문자열만 포함해야 합니다."
            assert table_name.startswith(f"{api_type}-profile"), f"`tables`의 테이블 이름은 반드시 {api_type}-profile로 시작해야 합니다."

        data = {"profile_id": profile_id, "tables": tables}
        if main_table:
            data["main_table"] = main_table

        return Profile(**self._request(method="POST", url=f"{MLS_PROFILES_API_URL}/{api_type}", data=data).results)

    def list_profiles(self, api_type: str, **kwargs) -> List[Profile]:
        """
        프로파일 리스트를 가져옵니다.

        ## Args

        - api_type: (str) 프로파일 타입 (`user`|`item`)
        - kwargs: (optional) (dict) 필터 조건
            - id: (int) 프로파일 고유 ID
            - profile_id: (str) 프로파일 이름
            - query: (str) 검색 문자
            - page: (int) 페이지 번호

        ## Returns
        list(`sktmls.profiles.Profile`)

        - id: (int) 프로파일 고유 ID
        - api_type: (str) 프로파일 타입 (`user`|`item`)
        - profile_id: (str) 프로파일 이름
        - tables: (list) 프로파일에 연결할 DynamoDB 테이블 이름 리스트
        - main_table: (str) 연결 테이블 중 기준이 되는 테이블 (추후 로직 추가 예정)
        - user: (str) 프로파일 소유 계정명

        ## Example

        ```python
        all_my_user_profiles = profile_client.list_profiles(api_type="user")
        ```
        """
        assert api_type in ["user", "item"], "api_type은 `user` 또는 `item`이어야 합니다."

        response = self._request(method="GET", url=f"{MLS_PROFILES_API_URL}/{api_type}", params=kwargs).results
        return [Profile(**profile) for profile in response]

    def get_profile(self, api_type: str, id: int = None, profile_id: str = None) -> Profile:
        """
        프로파일 정보를 가져옵니다.

        ## Args: `id` 또는 `profile_id` 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - api_type: (str) 프로파일 타입 (`user`|`item`)
        - id: (int) 프로파일 고유 ID
        - profile_id: (str) 프로파일 이름

        ## Returns
        `sktmls.clients.Client`

        - id: (int) 프로파일 고유 ID
        - api_type: (str) 프로파일 타입 (`user`|`item`)
        - profile_id: (str) 프로파일 이름
        - tables: (list) 프로파일에 연결할 DynamoDB 테이블 이름 리스트
        - main_table: (str) 연결 테이블 중 기준이 되는 테이블 (추후 로직 추가 예정)
        - user: (str) 프로파일 소유 계정명

        ## Example

        ```python
        profile_by_id = profile_client.get_profile(api_type="user", id=3)
        profile_by_profile_id = profile_client.get_profile(api_type="item", profile_id="my_item_profile")
        ```
        """
        assert api_type in ["user", "item"], "api_type은 `user` 또는 `item`이어야 합니다."
        assert id or profile_id, "`id` 또는 `profile_id` 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        profiles = self.list_profiles(api_type=api_type, id=id, profile_id=profile_id)
        if len(profiles) == 0:
            raise MLSClientError(code=404, msg="프로파일이 없습니다.")
        if len(profiles) > 1:
            raise MLSClientError(code=409, msg="프로파일이 여러개 존재합니다.")
        return profiles[0]

    def update_profile(
        self, profile: Profile, profile_id: str = None, tables: List[str] = None, main_table: str = None
    ) -> Profile:
        """
        프로파일 정보를 수정합니다.

        ## Args

        - profile_id: (optional) (str) 프로파일 이름
        - tables: (optional) (list) 프로파일에 연결할 DynamoDB 테이블 이름 리스트
            - api_type이 `user`인 경우 모든 테이블 이름은 `user-profile`로 시작해야 합니다.
            - api_type이 `item`인 경우 모든 테이블 이름은 `item-profile`로 시작해야 합니다.
            - 테이블 간 중복 키에 대한 값 우선 순위는 왼쪽 테이블이 오른쪽 테이블보다 높습니다.
            - 연결된 모든 테이블 내에 없는 키에 대한 값은 null로 반환됩니다.
        - main_table: (optional) (str) 서비스 대상 여부를 결정하는 테이블
            - 기준 테이블에 user_id/item_id가 존재하지 않는 경우 다른 테이블에 데이터가 있다 하더라도 서비스되지 않습니다.
            - 기준 테이블에 user_id/item_id가 존재하는 경우 다른 테이블의 데이터 여부와 관계 없이 반드시 서비스됩니다.
            - 미선택 시 `tables` 파라미터에 명시된 테이블 중 첫 번째 테이블로 설정됩니다.

        ## Returns
        `sktmls.profiles.Profile`

        - id: (int) 프로파일 고유 ID
        - api_type: (str) 프로파일 타입 (`user`|`item`)
        - profile_id: (str) 프로파일 이름
        - tables: (list) 프로파일에 연결할 DynamoDB 테이블 이름 리스트
        - main_table: (str) 연결 테이블 중 기준이 되는 테이블 (추후 로직 추가 예정)
        - user: (str) 프로파일 소유 계정명

        ## Example

        ```python
        profile = profile_client.profile_profile(
            profile=my_profile,
            profile_id="updated_profile",
            tables=["user-profile", "user-profile-daily"]
        )
        ```
        """
        assert type(profile) == Profile
        assert profile.api_type in ["user", "item"], "프로파일의 api_type은 `user` 또는 `item`이어야 합니다."

        data = profile.get()
        if profile_id:
            data["profile_id"] = profile_id
        if tables:
            assert type(tables) == list
            for table_name in tables:
                assert type(table_name) == str, "`tables`는 문자열만 포함해야 합니다."
                assert table_name.startswith(
                    f"{profile.api_type}-profile"
                ), f"`tables`의 테이블 이름은 반드시 {profile.api_type}-profile로 시작해야 합니다."
            data["tables"] = tables
        if main_table:
            data["main_table"] = main_table

        return profile.reset(
            **self._request(
                method="PUT", url=f"{MLS_PROFILES_API_URL}/{profile.api_type}/{profile.id}", data=data
            ).results
        )

    def delete_profile(self, profile: Profile) -> MLSResponse:
        """
        프로파일을 삭제합니다.

        ## Args

        - profile: (`sktmls.profiles.Profile`) 프로파일 객체

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```python
        profile_client.delete_profile(my_profile)
        ```
        """
        assert type(profile) == Profile
        assert profile.api_type in ["user", "item"], "프로파일의 api_type은 `user` 또는 `item`이어야 합니다."

        return self._request(method="DELETE", url=f"{MLS_PROFILES_API_URL}/{profile.api_type}/{profile.id}")
